%----------------------------------------------------------------------
% auxiliary.pl
%
% version 2.0, 2003-2008
%
% Linh Anh Nguyen 
% Institute of Informatics 
% Warsaw University
% nguyen@mimuw.edu.pl
%
%----------------------------------------------------------------------

%======================================================================
% set_option(Name, Value)
% get_option(Name, Value)
% set_option(Name, Parameter, Value)
% get_option(Name, Parameter, Value)
% set_info(Name, Value)
% get_info(Name, Value)
%
% variant(Term1, Term2)
% ground_form(Term1, Term2)
% ground_form_ma(Term1, Term2) : ground form of a modal atom
% termDepth(Term, TermDepth)
%
% check_calculus(Cal, Family)
% bound_index(I, N)
% print_msgs(Severity, Messages)
% check(Formula, Messages)
% check_not(Formula, Messages)
% forall(Index, Range, Instruction)
% for(Index, Range, Instruction)
% to_list(CommaSepList, List)
% from_list(List, CommaSepList)
%
% ex_modality(Cal, Modality) - existential modality 
% un_modality(Cal, Modality) - universal modality 
% make_dual_modality(Cal, Modality, DualModality)
% make_blf_modality(Cal, Modality, BoxLiftingForm)
% modality_instance(Cal, Instance, Modality)
% pre_check_modality_instance(Cal, Instance, Modality)
% post_check_modality_instance(Cal, Instance, Modality)
% almost_normal_labeled_form(Cal, Modality)
% pre_check_almost_normal_labeled_form(Cal, Modality)
% post_check_almost_normal_labeled_form(Cal, Modality)
%
% break_modal_atom(ModalAtom, Modality, ClassicalAtom)
% break_modality(Modality, PrefixLength, Prefix, Suffix)
%   Description: Modality and PrefixLength are inputs.
% break_rule_body(RuleBody, PreCondition, ModalAtom, RemainingFormula)
% get_pred(Atom, Predicate)
%
% my_random_permutation(ListIn, ListOut)
% slight_random_permutation(ListIn, ListOut)
%======================================================================

:- dynamic option/2, option/3, info/2.

set_option(Name, Value) :-
	atomic(Name), 
	retractall(option(Name, _)),
	assert(option(Name, Value)).

get_option(Name, Value) :- 
	option(Name, Value), !.

set_option(Name, Param, Value) :-
	atomic(Name), atomic(Param), 
	retractall(option(Name, Param, _)),
	assert(option(Name, Param, Value)).

get_option(Name, Param, Value) :- 
	(atomic(Name), atomic(Param)) -> 
		once(option(Name, Param, Value));
		option(Name, Param, Value).

set_info(Name, Value) :-
	atomic(Name), 
	retractall(info(Name, _)),
	assert(info(Name, Value)).

get_info(Name, Value) :- 
	info(Name, Value), !.

%======================================================================

% num2name(K, X): use x__ as prefix and K as suffix
num2name(K, X) :- name(K, L), name(X, [120,95,95|L]). 

subst(X/T, [Y/T|_]) :- X == Y, !.
subst(X/T, [_|L]) :- subst(X/T, L).

ground_form(T,T2) :- gform(T,1,[],T2,_,_).
gform(T,K,S,T,K,S) :- atomic(T), !.
gform(T,K,S,T2,K,S) :- var(T), subst(T/T2,S), !. 
gform(T,K,S,T2,K2,S2) :- var(T), !, num2name(K,T2), K2 is K+1, S2 = [T/T2|S]. 
gform(T,K,S,T2,K2,S2) :- T =.. [F|L], gform2(L,K,S,L2,K2,S2), T2 =.. [F|L2].
gform2([],K,S,[],K,S).
gform2([X|L],K,S,[X2|L2],K2,S2) :- 
	gform(X,K,S,X2,K3,S3), 
	gform2(L,K3,S3,L2,K2,S2).

variant(S,T) :- ground_form(S,F), ground_form(T,F).

ground_form_ma(M : A, (A2,M2)) :- ground_form(M : A, M2 : A2). 
	% just for efficiency of "tabled"/4

termDepth(T, 0) :- var(T), !. 
termDepth(T, 0) :- atomic(T), !. 
termDepth([], 1) :- !. 
termDepth([X | L], K) :- !, 
	termDepth([X | L], 0, K2),
	K is K2 + 1.
termDepth(T, K) :- 
	T =.. [_ | L],
	termDepth(L, 0, K2),
	K is K2 + 1.
termDepth([], A, A).
termDepth([X | L], A, K) :-
	termDepth(X, K2), 
	(K2 > A -> A2 = K2; A2 = A),
	termDepth(L, A2, K).

%======================================================================

check_calculus(Cal, Family) :- once(is_calculus(Cal, Family)).

% increasingly :
bound_index(I, N) :- 
	number(I) -> I =< N; bound_index(I, 1, N).
bound_index(I, K, N) :- 
        I = K; (K < N, K2 is K+1, bound_index(I, K2, N)).
/*
% decreasingly :
bound_index(I, N) :- 
	number(I) -> I =< N; bound_index(I, N, N).
bound_index(I, K, N) :- 
        I = K; (K > 1, K2 is K-1, bound_index(I, K2, N)).
*/

print_msgs([Msg]) :- !, write(Msg).
print_msgs([Msg | L]) :- !, write(Msg), print_msgs(L).
print_msgs(Msg) :- write(Msg).

print_msgs(Severity, Msgs) :-
	Severity = error -> write('! '), print_msgs(Msgs), nl; 
	Severity = debug -> 
	  (get_option(debug, true) -> 
		  write('% '), print_msgs(Msgs), nl; 
		  true); 
	get_option(quiet, true) -> true;
	(Severity = warning -> write('* ');
	 Severity = informational -> write('% ');
 	 Severity = info -> write('% ');
	 write(''' ')),
	print_msgs(Msgs), nl.

check(Formula, Messages) :-
	Formula -> true; print_msgs(error, Messages), fail.

check_not(Formula, Messages) :-
	Formula -> (print_msgs(error, Messages), fail); true.

forall(Index, Range, Inst) :- 
	\+ ( member(Index, Range), \+ call(Inst) ).
 
for(Index, Range, Inst) :-
	member(Index, Range), call(Inst), fail.
for(_, _, _).

to_list(X, _) :- var(X), !, fail.
to_list((X,Y), [X|List]) :- !, to_list(Y, List).
to_list(X, [X]).

from_list([X,Y|List], (X,Z)) :- !, from_list([Y|List], Z).
from_list([X], X).
from_list([], true).

%======================================================================

ex_modality(_, []) :- !.

ex_modality(Cal, [O|M]) :-
	\+ universal_modal_operator(Cal, O), !, 
	ex_modality(Cal, M). 

%----------------------------------------------------------------------

un_modality(_, []) :- !.

un_modality(Cal, [O|M]) :-
	universal_modal_operator(Cal, O), !, 
	un_modality(Cal, M). 

%----------------------------------------------------------------------

make_dual_modality(_, [], []) :- !.

make_dual_modality(Cal, [O|M], [DO|DM]) :- 
	dual_modal_operator(Cal, O, DO), !, 
	make_dual_modality(Cal, M, DM). 

%----------------------------------------------------------------------

make_blf_modality(_, [], []) :- !.

make_blf_modality(Cal, [O|M], [O2|M2]) :- 
	box_lifting_form(Cal, O, O2), !, 
	make_blf_modality(Cal, M, M2). 

%----------------------------------------------------------------------

modality_instance(_, [], []).

modality_instance(Cal, [OI|MI], [O|M]) :-
	operator_instance(Cal, OI, O), 
	modality_instance(Cal, MI, M).

pre_check_modality_instance(_, [], []).

pre_check_modality_instance(Cal, [OI|MI], [O|M]) :-
	pre_check_operator_instance(Cal, OI, O), 
	pre_check_modality_instance(Cal, MI, M).

post_check_modality_instance(_, [], []).

post_check_modality_instance(Cal, [OI|MI], [O|M]) :-
	post_check_operator_instance(Cal, OI, O), 
	post_check_modality_instance(Cal, MI, M).

%----------------------------------------------------------------------

almost_normal_labeled_form(_, []) :- !.
almost_normal_labeled_form(_, [_]) :- !.
almost_normal_labeled_form(Cal, Modality) :- !, 
	append(M, [_], Modality), 
	normal_labeled_form(Cal, M).
	
pre_check_almost_normal_labeled_form(_, []) :- !.
pre_check_almost_normal_labeled_form(_, [_]) :- !.
pre_check_almost_normal_labeled_form(Cal, Modality) :- !, 
	append(M, [_], Modality), 
	pre_check_normal_labeled_form(Cal, M).
	
post_check_almost_normal_labeled_form(_, []) :- !.
post_check_almost_normal_labeled_form(_, [_]) :- !.
post_check_almost_normal_labeled_form(Cal, Modality) :- !, 
	append(M, [_], Modality), 
	post_check_normal_labeled_form(Cal, M).
	
%======================================================================

break_modal_atom(A, _, _) :- 
	var(A), !, 
	print_msgs(error, ['Encountered an atom in the form ', A]),
	fail.
 
break_modal_atom(M : A, M2, A2) :- !, 
	break_modal_atom(A, M3, A2),  
	(M3 = [] -> M2 = M; append(M, M3, M2)).

break_modal_atom(A, [], A).
	
%----------------------------------------------------------------------
% break_modality(Modality, PrefixLength, Prefix, Suffix)

break_modality(M, L, P, S) :-
	length(M, L2),
	(L2 =:= L -> P = M, S = [];
	 L2 =:= L+1 -> S = [_], append(P, S, M)),
   	!.
   
break_modality(M, 0, [], M) :- !.

break_modality([O|M], K, [O|M2], M3) :-
	K > 0, K2 is K -1,
	break_modality(M, K2, M2, M3).

%----------------------------------------------------------------------
% break_rule_body(RuleBody, PreCondition, ModalAtom, RemainingFormula)

break_rule_body(M : Atom, true, M : Atom, true) :- !.

break_rule_body((M : Atom, F), true, M : Atom, F) :- !.

break_rule_body((F1, F2), (F1, PreCond), Ao, RF) :-
	break_rule_body(F2, PreCond, Ao, RF). 

%----------------------------------------------------------------------

get_pred(Atom, _) :- 
	var(Atom), !, 
	print_msgs(error, ['Encountered an atom in the form ', Atom]),
	fail.

get_pred(_ : Atom, Pred) :- 
	!, get_pred(Atom, Pred).

get_pred(CAtom, Pred/Arity) :- 
	functor(CAtom, Pred, Arity).

%======================================================================

my_random_permutation(ListIn, ListOut) :- 
	random_permutation(ListIn, ListOut, fail).

slight_random_permutation(ListIn, ListOut) :- 
	random_permutation(ListIn, ListOut, true).

random_permutation(ListIn, ListOut, Slight) :-
	length(ListIn, N), 
	random_permutation(ListIn, N, [], ListOut, Slight).

random_permutation([], _, Acc, Acc, _) :- !.

random_permutation(ListIn, N, Acc, ListOut, Slight) :- 
	N2 is N+1, 
	(Slight ->  
	   (random(1, N2, A), random(1, N2, B), 
	    (A < B -> K = B; K = A));
    	 random(1, N2, K)), 
	nth1(K, ListIn, Elem), select(Elem, ListIn, Rest), !, 
	N3 is N-1,
	random_permutation(Rest, N3, [Elem | Acc], ListOut, Slight).

%======================================================================

