%------------------------------------------------------------------------
% datastr.pl: Data structures
%
% version 2.0, 2003-2008
%
% Linh Anh Nguyen 
% Institute of Informatics
% Warsaw University
% nguyen@mimuw.edu.pl
%
%------------------------------------------------------------------------

%========================================================================
% Declaration of calculi, modal predicates, and dum (defined using modal 
% atoms) predicates.
%
% Predicates:
%   init_dec
%
%   dec_calculus(ModalCal)
%   get_list_of_calculi(ListOfCalculi)
%   
%   dec_pred(Kind, Predicate)
%   dec_preds(Kind, Predicates)
%   is_pred_of_kind(Kind, Predicate)
%   get_list_of_preds(Kind, ListOfPredicates)
%
% Data structure:
%   list_of_calculi(ListOfCalculi)
%   list_of_preds(Kind, ListOfPredicates)
%------------------------------------------------------------------------

:- dynamic list_of_calculi/1, list_of_preds/2.

init_dec :-
	retractall(list_of_calculi(_)),
	retractall(list_of_preds(modal, _)),
	retractall(list_of_preds(dum, _)),
	assert(list_of_calculi([classical])),
	assert(list_of_preds(modal, [])),
	assert(list_of_preds(dum, [])).

dec_calculus(Cal) :- 
	atomic(Cal), 
	get_list_of_calculi(Calculi), 
	retractall(list_of_calculi(_)),
	assert(list_of_calculi([Cal | Calculi])),
	findall((Name, Value), 
	        (member(Name, [limit_modality_length, limit_rSat, limit_rNF, 
	                       use_calling_history, check_in_two_steps, 
			       copiable]), 
                 get_option(Name, Value)),
	        OList), 
	for((Name, Value), OList, 
	    (get_option(Name, Cal, _) -> true; set_option(Name, Cal, Value))).

get_list_of_calculi(Calculi) :- 
	list_of_calculi(Calculi) -> true; Calculi = [].

pred_declared(Pred) :- 
	get_list_of_preds(_, Preds), 
	memberchk(Pred, Preds).

dec_pred(Kind, Pred) :- 
	atomic(Kind), ground(Pred), 
	check_not(pred_declared(Pred), ['Predicate ', Pred, ' already declared.']), 
	get_list_of_preds(Kind, Preds), 
	retractall(list_of_preds(Kind, _)),
	assert(list_of_preds(Kind, [Pred | Preds])).

dec_preds(_, []).
dec_preds(Kind, [Pred | L]) :- 
	dec_pred(Kind, Pred),
	dec_preds(Kind, L).

get_list_of_preds(Kind, Preds) :-
	list_of_preds(Kind, Preds) -> true; Preds = [].

is_pred_of_kind(Kind, Pred) :- 
	atomic(Kind), 
	get_list_of_preds(Kind, Preds), 
	memberchk(Pred, Preds).

%========================================================================
% History of called rules (in the current resolution cycle)
%
% Predicates:
%   init_calling_history(Category, Calculus, Atom)
%   log_called_rule(Calculus, Category, RuleName)
%   get_calling_history(Category, Calculus, Atom, ListOfCalledRuleNames)
%
% Data structure:
%   calling_history(Category, Calculus, Atom, ListOfCalledRuleNames)
%     Description: Atom is the atom that was given for solving by 
%     the rules in ListOfCalledRuleNames (in the reverse order).
%     There is at most one clause calling_history(Category, _, _, _)
%     for each Catergory. ListOfCalledRuleNames contains names of 
%     called rules.
%------------------------------------------------------------------------

:- dynamic calling_history/4.

init_calling_history(Cat, Cal, Atom) :- 
	get_option(use_calling_history, Cal, false) -> true; 
	atomic(Cat), atomic(Cal), 
	(calling_history(Cat, Cal, Atom2, CRNames2) -> true;
	 (Atom2 = nil, CRNames2 = [])), 
	((retractall(calling_history(Cat, Cal, _, _)), 
	  assertz(calling_history(Cat, Cal, Atom, [])));
	 (retractall(calling_history(Cat, Cal, _, _)), 
	  (Atom2 == nil -> true; 
	   assertz(calling_history(Cat, Cal, Atom2, CRNames2))),
	   fail)).
	
log_called_rule(Cal, Cat, RName) :- 
	get_option(use_calling_history, Cal, false) -> true; 
	calling_history(Cat, Cal, Atom, CRNames), 
	((retractall(calling_history(Cat, Cal, _, _)),
	  assertz(calling_history(Cat, Cal, Atom, [RName | CRNames])));
	 (retractall(calling_history(Cat, Cal, _, _)),
	  assertz(calling_history(Cat, Cal, Atom, CRNames)),
	  fail)).
	    
get_calling_history(Cat, Cal, Atom, CRNames) :- 
	calling_history(Cat, Cal, Atom, CRNames), !.

%========================================================================
% Rules
%
% Predicates:
%   rule_cats([pre_rSat, rSat, post_rSat, rNF])
%   gen_rule_name(Name) 
%   dec_mrule(Name, Rule)
%   get_mrule(Name, Rule)
%     Description: Name is input, Rule is output
%   get_mrule2(Name, Head, PreCond, ModAtom, PostComp)
%   set_list_of_mrules(Calculus, Category, ListOfRuleNames)
%   get_list_of_mrules(Calculus, Category, ListOfRuleNames)
%     Description: Calculus and Category are inputs, 
%	           ListOfRuleNames is output 
%   assertz_mrule(Calculus, Category, RuleName) 
%
% Data structure:
%   mrule(Name, Rule, Head, PreCond, ModAtom, PostComp)
%   list_of_mrules(Cal, Category, ListOfRuleNames)
%     Description: Category is pre_rSat, rSat, post_rSat, or rNF
%------------------------------------------------------------------------

:- dynamic mrule/6, list_of_mrules/3.

rule_cats([pre_rSat, rSat, post_rSat, rNF]).

gen_rule_name(Name) :- 
	findall(RName, mrule(RName, _, _, _, _, _), RNames), 
	gen_rule_name(Name, 1, RNames). 
gen_rule_name(N, N, RNames) :- 
	\+ memberchk(N, RNames), !.
gen_rule_name(Name, N, RNames) :- 
	N2 is N+1,
	gen_rule_name(Name, N2, RNames).

dec_mrule(Name, Rule) :-
	atomic(Name), 
	(mrule(Name, _, _, _, _, _) -> 
	   print_msgs(error, ['Rule ', Name, ' not REDEFINED.']);
	   (Rule = (Head :- Body),
	    break_rule_body(Body, PreCond, ModAtom, PostComp), 
	    assert(mrule(Name, Rule, Head, PreCond, ModAtom, PostComp)))).

get_mrule(Name, Rule) :- mrule(Name, Rule, _, _, _, _), !.

get_mrule2(Name, Head, PreCond, ModAtom, PostComp) :-
	mrule(Name, _, Head, PreCond, ModAtom, PostComp), !.

set_list_of_mrules(Cal, Cat, RNames) :- 
	atomic(Cal), atomic(Cat), 
	get_list_of_calculi(Calculi),
	(member(Cal, Calculi) -> true; dec_calculus(Cal)),
	retractall(list_of_mrules(Cal, Cat, _)), 
	assert(list_of_mrules(Cal, Cat, RNames)).

get_list_of_mrules(Cal, Cat, RNames) :- 
	list_of_mrules(Cal, Cat, RNames) -> true; RNames = [].

assertz_mrule(Cal, Cat, RName) :- 
	atomic(Cal), atomic(Cat), 
	get_list_of_mrules(Cal, Cat, RNames0), 
	append(RNames0, [RName], RNames), 
	set_list_of_mrules(Cal, Cat, RNames). 

%========================================================================
% Modal clauses
%
% Predicates:
%   break_clause(Cal, Clause, Context, Head, Body)
%   compose_clause(Context, Head, Body, Clause)
%   gen_clause_name(Name) 
%   dec_mclause(Name, Context, Head, Body)
%   get_mclause(Name, Context, Head, Body)
%   get_mclause(Name, Context, Head, Body, BodyLength)
%   set_list_of_mclauses(Calculus, Predicate, ListOfClauseNames)
%   get_list_of_mclauses(Calculus, Predicate, ListOfClauseNames)
%   asserta_mclause(Calculus, Predicate, ClauseName)
%   assertz_mclause(Calculus, Predicate, ClauseName)
%   assert_mclause(Method, Calculus, Predicate, ClauseName)
%	Method = s --> sort by body lengths of clauses
%
% Data structure:
%   mclause(Name, Context, Head, Body, BodyLength)
%   list_of_mclauses(Calculus, Predicate, ListOfClauseNames)
%------------------------------------------------------------------------

:- dynamic mclause/5, list_of_mclauses/3.

extract_context(Cal, Modality, Context, Tail) :-
	extract_context(Cal, Modality, [], Context, Tail).

extract_context(_, [], Acc, Acc, []) :- !.

extract_context(Cal, [O|M], Acc, Acc, [O|M]) :- 
	(\+ universal_modal_operator(Cal, O); 
	 (append(Acc, [O], Acc2), 
	  (get_option(check_in_two_steps, Cal, true) ->	 
	     \+ pre_check_normal_labeled_form(Cal, Acc2);
	     \+ normal_labeled_form(Cal, Acc2)))
	), !.
	
extract_context(Cal, [O|M], Acc, Context, Tail) :- 
	append(Acc, [O], Acc2), 
	extract_context(Cal, M, Acc2, Context, Tail). 

break_clause(Cal, Clause, Context, Head, Body) :- 
	Clause = (Context : (Head :- Body)) -> true; 
	Clause = (Head :- Body) -> Context = [];
	Clause = (Context : M : H) -> (Head = M : H, Body = true);  
	Body = true, 
	break_modal_atom(Clause, M, A), 
	extract_context(Cal, M, Context, M2),
	(M2 = [] -> Head = A; Head = M2 : A).

compose_clause(Context, Head, Body, Clause) :- 
	(Body = true, Context = []) -> Clause = Head; 
	Body = true -> Clause = Context : Head; 
	Context = [] -> Clause = (Head :- Body); 
	Clause = Context : (Head :- Body).

gen_clause_name(Name) :- 
	findall(CName, mclause(CName, _, _, _, _), CNames), 
	gen_clause_name(Name, 1, CNames). 
gen_clause_name(N, N, CNames) :- 
	\+ member(N, CNames), !.
gen_clause_name(Name, N, CNames) :- 
	N2 is N+1,
	gen_clause_name(Name, N2, CNames).

dec_mclause(Name, Context, Head, Body) :-
	atomic(Name), 
	retractall(mclause(Name, _, _, _, _)),
	(Body = true -> BodyLength = 0; 
	 (to_list(Body, Body2), length(Body2, BodyLength))), 
	assert(mclause(Name, Context, Head, Body, BodyLength)).

get_mclause(Name, Context, Head, Body) :-
	mclause(Name, Context, Head, Body, _), !.

get_mclause(Name, Context, Head, Body, BodyLength) :-
	mclause(Name, Context, Head, Body, BodyLength), !.

set_list_of_mclauses(Cal, Pred, CNames) :- 
	atomic(Cal), ground(Pred), 
	retractall(list_of_mclauses(Cal, Pred, _)),
	assert(list_of_mclauses(Cal, Pred, CNames)).

get_list_of_mclauses(Cal, Pred, CNames) :- 
	list_of_mclauses(Cal, Pred, CNames) -> true; CNames = [].

asserta_mclause(Cal, Pred, CName) :- 
	assert_mclause(a, Cal, Pred, CName). 

assertz_mclause(Cal, Pred, CName) :- 
	assert_mclause(z, Cal, Pred, CName). 

assert_mclause(Method, Cal, Pred, CName) :- 
	get_list_of_preds(modal, ModalPreds),
	get_list_of_preds(dum, DumPreds),
	((memberchk(Pred, ModalPreds); memberchk(Pred, DumPreds)) -> true; 
	   dec_pred(modal, Pred)), 

	get_list_of_mclauses(Cal, Pred, CNames),
	insert(Method, CName, CNames, CNames2), 
	set_list_of_mclauses(Cal, Pred, CNames2).
 
insert(a, CName, CNames, [CName | CNames]). 

insert(z, CName, CNames, CNames2) :- 
	append(CNames, [CName], CNames2).  

insert(s, CName, CNames, CNames2) :- 
	mclause(CName, _, _, _, BodyLength), 
	insert_s(CName, BodyLength, CNames, CNames2). 
	
insert_s(CName, _, [], [CName]).

insert_s(CName, Length, [CName2 | CNames], CNames2) :- 
	mclause(CName2, _, _, _, Length2), 
	(Length < Length2 ->
	   CNames2 = [CName, CName2 | CNames];
	   (insert_s(CName, Length, CNames, CNames3), 
	    CNames2 = [CName2 | CNames3])).	   

%========================================================================
% Tabulation for modal atoms
%
% Predicates:
%   get_tabulation_size(K)
%   table_answer(Cal, A, B0)
%   reset_tabulation
%   delete_tabulation
%
% Data structure:
%   called(A, Cal) : a modal atom A0 s.t. ground_form_ma(A0,A) has been
%	called in calculus Cal.
%   tabled(A, Cal, B, B0) : B0 is an answer instance of A0 in calculus 
%	Cal, where ground_form_ma(A0,A) and ground_form_ma(B0,B) hold.
%	There is at most one B0 for the tuple (A, Cal, B).
%   locally_completed(A, Cal) : all answers for A0 in Cal, where 
%	ground_form_ma(A0,A) holds, have been "tabled" using tabled/4.
%   tabulation_size(K) : K is the number of tabled modal atoms.
%------------------------------------------------------------------------

:- dynamic locally_completed/2, called/2, tabled/4, tabulation_size/1.

get_tabulation_size(K) :- 
	tabulation_size(K) -> true; K = 0.

table_answer(Cal, A, B0) :-
	ground_form_ma(B0, B),
	(tabled(A, Cal, B, _) -> true;
	 assertz(tabled(A, Cal, B, B0)),
	 get_tabulation_size(K),
	 K2 is K + 1,
	 retractall(tabulation_size(_)),
	 assert(tabulation_size(K2))).

reset_tabulation :- 
	retractall(called(_, _)),
	retractall(locally_completed(_, _)).
	
delete_tabulation :- 
	retractall(called(_, _)),
	retractall(locally_completed(_, _)),
	retractall(tabled(_, _, _, _)),
	retractall(tabulation_size(_)).

%========================================================================
% Create a copy of a calculus
%
% is_calculus(Calculus, Family)
% copy_calculus(SrcCal, DestCal)
%    Description: This predicate creates DestCal from SrcCal by
%    copying mrules and options. This is allowed if 
%    get_option(copiable, SrcCal, true) holds. Required predicates like 
%    universal_modal_operator(Cal, Oper) must have been writen for
%    SrcCal so that they are not needed to be rewriten for DestCal.
%------------------------------------------------------------------------

is_calculus(Cal, Cal).

copy_calculus(SrcCal, DestCal) :-
	atomic(SrcCal), atomic(DestCal), 
	get_option(copiable, SrcCal, true),

	get_list_of_calculi(Calculi), 
	retractall(list_of_calculi(_)),
	assert(list_of_calculi([DestCal | Calculi])),

	rule_cats(Cats),
	for(Cat, Cats, 
	  (get_list_of_mrules(SrcCal, Cat, RNames), 
	   set_list_of_mrules(DestCal, Cat, RNames))),

	findall((OName, Value), option(OName, SrcCal, Value), OList),
	for((OName, Value), OList, set_option(OName, DestCal, Value)),
	set_option(copiable, DestCal, false),
	assertz(is_calculus(DestCal, SrcCal)).

