%------------------------------------------------------------------------
% mprolog.pl
%
% version 2.0, 2003-2008
%
% Linh Anh Nguyen, 
% Institute of Informatics 
% Warsaw University
% nguyen@mimuw.edu.pl
%
%------------------------------------------------------------------------
%
% Consulting modal calculi:
%
% 	consult_calculi(Files)
%
% Consulting programs:
%
% 	mconsult(Files, Calculus)
% 	mconsult(Files)
%
% Restoring program to the status before the last consulting:
%
% 	mrestore
%
% Calling goals:
%
% 	mcall(Goal)
% 	mcall(Goal, Calculus)
%
% Iterative deepening search w.r.t. the number of applications of
% rSat/rNF rules:
%
% 	mcall2(Goal)
% 	mcall2(Goal, Calculus)
%
% Iterative deepening search w.r.t. proof size
%
% 	mcall3(Goal)
% 	mcall3(Goal, Calculus)
%
% Display calculi and definitions of predicates: 
%
% 	get_number_of_mclauses(Calculus, Predicate, Number)
% 	get_modal_clause(Calculus, Predicate, Number, Clause)
% 	mlisting(Calculus)
% 	mlisting(Predicate)
%
% Turn on/off trace mode:
%
%	mtrace
%	nomtrace
%
%------------------------------------------------------------------------

:- use_module(library(lists)).

:- include('platform.pl').
:- include('auxiliary.pl').
:- include('params.pl').
%:- include('tracer.pl').
:- include('datastr.pl').
:- include('parser.pl').
:- include('solver.pl').
:- include('compiler.pl').

:- init_dec.

% For SWI-Prolog, uncomment the following line:
%:- swi.

:- option(sicstus, false) -> true;
	nl, write('IMPORTANT FOR SWI-PROLOG: CALL swi/0 NOW!'), nl, nl.

%------------------------------------------------------------------------

msave :-
	save_program('msaved.tmp') -> set_info(msaved, 'msaved.tmp');

	% try to save the program in the home directory
	get_option(mprolog_home, HOME), 
	name(HOME, Dir),
	append(Dir, "/msaved.tmp", FName),
	name(FileName, FName),
	(save_program(FileName) -> 
		set_info(msaved, FileName);
 	 	set_info(msaved, null)).

mrestore :- 
	get_option(save_before_consulting, true), 
	get_info(msaved, FileName),
	FileName \= null, 
	restore(FileName).

%------------------------------------------------------------------------

consult_calculi(Files) :-
	(get_option(save_before_consulting, true) -> msave; true), 
	consult_cal(Files). % defined in parser.pl

mconsult(Files, Cal) :-
	(get_option(save_before_consulting, true) -> msave; true), 
	mconsult_prog(Files, Cal). % defined in parser.pl	

mconsult(Files) :-
	mconsult(Files, classical).

%------------------------------------------------------------------------

:- dynamic answer/1.

mcall(Goal) :- 
	get_option(current_calculus, Cal), 
	print_msgs(info, ['Calling in ', Cal, '.']), 
	mcall(Goal, Cal).

mcall(Goal, Cal) :- 
	retractall(answer(_)),
	get_option(limit_rule_applications, LimitRA), 
	set_option(max_rule_applications, LimitRA), 
	get_option(limit_proof_size, LimitPS), 
	set_option(max_proof_size, LimitPS), 
	mcall_(Cal, Goal).

%------------------------------------------------------------------------

mcall2(Goal) :- 
	get_option(current_calculus, Cal), 
	print_msgs(info, ['Calling in ', Cal, '.']), 
	mcall2(Goal, Cal).

mcall2(Goal, Cal) :- 
	retractall(answer(_)),
	get_option(limit_proof_size, LimitPS), 
	set_option(max_proof_size, LimitPS), 
	get_option(limit_rule_applications, LimitRA), 
	get_option(deepening_step_for_rule_applications, DeepeningStep), 
	mcall2(Goal, Cal, DeepeningStep, DeepeningStep, LimitRA).

mcall2(Goal, Cal, CurrentLimit, DeepeningStep, Limit0) :- 
	CurrentLimit =< Limit0,
	(set_option(max_rule_applications, CurrentLimit), 
	 mcall_(Cal, Goal)
	 ;
	 CurrentLimit2 is CurrentLimit + DeepeningStep,
	 mcall2(Goal, Cal, CurrentLimit2, DeepeningStep, Limit0)
	).

%------------------------------------------------------------------------

mcall3(Goal) :- 
	get_option(current_calculus, Cal), 
	print_msgs(info, ['Calling in ', Cal, '.']), 
	mcall3(Goal, Cal).

mcall3(Goal, Cal) :- 
	retractall(answer(_)),
	get_option(limit_rule_applications, LimitRA), 
	set_option(max_rule_applications, LimitRA), 
	get_option(limit_proof_size, LimitPS), 
	get_option(deepening_step_for_proof_size, DeepeningStep), 
	mcall3(Goal, Cal, DeepeningStep, DeepeningStep, LimitPS).

mcall3(Goal, Cal, CurrentLimit, DeepeningStep, Limit0) :- 
	CurrentLimit =< Limit0,
	(set_option(max_proof_size, CurrentLimit), 
	 mcall_(Cal, Goal)
	 ;
	 CurrentLimit2 is CurrentLimit + DeepeningStep,
	 mcall3(Goal, Cal, CurrentLimit2, DeepeningStep, Limit0)
	).

%------------------------------------------------------------------------

get_number_of_mclauses(Cal, Pred, N) :- 
	get_list_of_mclauses(Cal, Pred, CNames), 
	length(CNames, N).

get_modal_clause(Cal, Pred, N, Clause) :- 
	get_list_of_mclauses(Cal, Pred, CNames), 
	nth1(N, CNames, Name), 
	get_mclause(Name, Context, Head, Body),
	compose_clause(Context, Head, Body, Clause).

mlisting(Cal) :- 
	atomic(Cal), 
	get_list_of_calculi(Calculi), 
	memberchk(Cal, Calculi), !, 
	write('OPTIONS: '), nl, 
	findall((Name, Value), get_option(Name, Cal, Value), OList),
	for((Name, Value), OList, 
	    (write(option(Name, Cal, Value)), nl)),
	write('RULES: '), nl, 
	rule_cats(Cats), 
	for(Cat, Cats, 
	    (get_list_of_mrules(Cal, Cat, RNames), 
	     (RNames = [] -> true;
	      write(Cat), write(' : '), nl, 
	      for(RName, RNames, 
	          (get_mrule(RName, Rule),
		     write((RName :: Rule)), nl))))),
	for(Kind, [modal, dum], 
	    ((Kind = modal -> write('MODAL CLAUSES: '); write('DUM CLAUSES: ')), 
	     nl, 
	     get_list_of_preds(Kind, Preds),
	     for(Pred, Preds, 
	         (get_list_of_mclauses(Cal, Pred, CNames), 
	          (CNames = [] -> true; 
	           for(CName, CNames, 
	               (get_mclause(CName, Context, Head, Body),
	                compose_clause(Context, Head, Body, Clause), 
	                write(Clause), write('.'), nl))))))).

mlisting(Pred) :- 
	atom(Pred), !, 
	(setof(Arity, (is_pred_of_kind(modal, Pred/Arity); 
                       is_pred_of_kind(dum, Pred/Arity)), Arities) ->
	   true;
	   print_msgs(error, [Pred, ' is not a modal/dum predicate.'])),
	for(A, Arities, mlisting2(Pred/A)).  

mlisting(Pred/Arity) :- 
	ground(Pred/Arity), 
	((\+ is_pred_of_kind(dum, Pred/Arity), 
	  \+ is_pred_of_kind(modal, Pred/Arity)) ->  
	 print_msgs(error, [Pred/Arity, ' is not a modal/dum predicate.']);  
	 mlisting2(Pred/Arity)).

mlisting2(Pred/Arity) :- 
	get_list_of_calculi(Calculi), 
	for(Cal, Calculi, 
	    (get_list_of_mclauses(Cal, Pred/Arity, CNames), 
	     (CNames = [] -> true;  
	      write(Cal), write(' : '), nl, 
	      for(CName, CNames, 
	          (get_mclause(CName, Context, Head, Body),
	           compose_clause(Context, Head, Body, Clause), 
	           write(Clause), write('.'), nl))))).

%------------------------------------------------------------------------

mtrace :- set_option(mtrace, true).
nomtrace :- set_option(mtrace, false).

%------------------------------------------------------------------------

