/*****************************************************************************
THE n WISE MEN PUZZLE: see wise_men_3.mpl for a description of the case n = 3. 
*****************************************************************************/

% Denote the wise men by numbers from 1 to n. A group of wise men is denoted
% by the ordered list consisting of the identifications of the men. The group 
% of all wise men [1,2,...,n] is held by the dynamic predicate all_wise_men/1 
% (because n is not fixed). A singleton group [X] can be treated as wise man X.

%-----------------------------------------------------------------------------
% Here are predicates required for groups.cal:

singleton_group([_]).

subgroup(G1, G2) :- subList(G1, G2), G1 \= G2, G1 = [_|_].

subList([], _).
subList([X], L) :- member(X, L).
subList([X, Y | L], L2) :- 
	member(X, L2), member(Y, L2), X < Y, 
	subList([Y | L], L2).

union_group(G1, G2, G3) :- merge(G1, G2, [ ], G3).

merge([X | L1], [Y | L2], Acc, L) :- X = Y, !, merge(L1, [Y | L2], Acc, L).
merge([X | L1], [Y | L2], Acc, L) :- X < Y, !, merge(L1, [Y | L2], [X | Acc], L).
merge([X | L1], [Y | L2], Acc, L) :- merge([X | L1], L2, [Y | Acc], L).
merge(L1, [], Acc, L) :- reverse(Acc, A), append(A, L1, L).
merge([], L2, Acc, L) :- reverse(Acc, A), append(A, L2, L).

%-----------------------------------------------------------------------------
% Here are auxiliary predicates to fix n and set up the group all_wise_men.
% Here we also set the limit_modality_length option to n.

set_number_of_men(N) :-
	generate_men(N, All), 
	set_info(all_wise_men, All), 
	set_option(limit_modality_length, cKD4Ig5a, N).
	
generate_men(N, All) :- 
	generate_men(N, [], All).
generate_men(0, Acc, Acc) :- !. 
generate_men(K, Acc, All) :-  
	K2 is K - 1, generate_men(K2, [K | Acc], All).

%=============================================================================
% Here are main predicates of the program.
% We use the predicate white(ListOfMen) to denote that at least one of the men
% in the list ListOfMen has a white card. We use black(X) to denote that wise 
% man X has a black card.

:- calculus cKD4Ig5a.

% At least one of the men has a white card.

[bel(All)]: (white(All) :- get_info(all_wise_men, All)).

% Let G2 be a group of men, X be a man in G2, and G = G2 - {X}.
% If at least one of the members of G2 has a white card and the card of X is 
% black, then at leat one the members of G has a white card.
  
[bel(All)]: (white(G) :- 
	get_info(all_wise_men, All), subList(G, All), black(X), 
	subList(G2, All), member(X, G2), delete(G2, X, G), white(G2)).

% If Y sits behinds all members of the group G, then white(G) if Y considers 
% that as possible.

[bel(All)]: (white(G) :- 
	get_info(all_wise_men, All), subList(G, All), append(_, [X], G), 
	member(Y, All), X < Y, [pos([Y])]:white(G)).

% Each of the wise men with a number different than 1 does not know the color 
% of his own card. In particular, each of those men considers that it is 
% possible that his own card is black.

[bel(All)]: ([pos([X])]:black(X) :- 
	get_info(all_wise_men, All), member(X, All), X \= 1).

% Change some default options:
:- set_option(limit_rSat, cKD4Ig5a, 1).
:- set_option(current_calculus, cKD4Ig5a).

%=============================================================================
% Results tested using MProlog 2.0, SICStus 4.0.2 on
% notebook Acer TravelMate 2482WXMi
% Intel(R) Celeron(R) M CPU 430 @ 1.73 GHz, 504 MB RAM 

% Does wise man 1 believe that his card is white? 
%    1s:   ?- set_number_of_men(10), mcall([bel([1])]:white([1])).
%    2s:   ?- set_number_of_men(11), mcall([bel([1])]:white([1])).
%    4s:   ?- set_number_of_men(12), mcall([bel([1])]:white([1])).
%    8s:   ?- set_number_of_men(13), mcall([bel([1])]:white([1])).
%   16s:   ?- set_number_of_men(14), mcall([bel([1])]:white([1])).

%-----------------------------------------------------------------------------

% It speeds up if the program is compiled as follows:
%  1) delete the inclusion of trace.pl from mprolog.pl
%  2) run prolog and change the working directory to the directory of mprolog
%  3) consult mprolog.pl
%  4) :- consult_calculi('groups.cal').
%  5) :- mconsult('wise_men_n.pl').
%  6) :- compile_loaded_program.
%  7) restart prolog 
%  8) consult compile_p.pl
%  9) change program.po to wise_men_n.po

%    4s:   ?- set_number_of_men(14), mcall([bel([1])]:white([1])).

